import { arrayLast } from "../../Common/ArrayExt";
import { Container } from "./Container";
import { NestFiler } from "../Common/Filer";
import { Part } from "./Part";
import { Path } from "./Path";
import { PlaceType } from "./PlaceType";
import { RandomIndex } from "../Common/Random";

/**
 * 个体(表示某一次优化的结果,或者还没开始优化的状态)
 * 个体是由一堆零件组成的，零件可以有不同的状态。
 *
 * 个体单独变异
 *  可以是某个零件的旋转状态发生改变
 *  可以是零件的顺序发生改变
 *
 * 个体交配(感觉暂时不需要)
 *  个体和其他个体进行基因交换
 */
export class Individual
{
    //强壮程度 越低越好
    Fitness: number;
    constructor(public Parts?: Part[], public mutationRate = 0.5) { }

    Clone()
    {
        let p = new Individual(this.Parts.map(p => p.Clone(), this.mutationRate));
        p.Mutate();
        return p;
    }

    /**
    * 突变
    */
    Mutate()
    {
        if (this.mutationRate > 0.5)
            for (let i = 0; i < this.Parts.length; i++)
            {
                let rand = Math.random();
                if (rand < this.mutationRate * 0.5)
                {
                    //和下一个调换顺序
                    let j = i + 1;
                    if (j < this.Parts.length)
                        [this.Parts[i], this.Parts[j]] = [this.Parts[j], this.Parts[i]];
                }
                if (rand < this.mutationRate)
                    this.Parts[i].Mutate();
            }
        else
        {
            //洗牌
            let rand = Math.random();
            if (rand < 0.5)
            {
                let index = RandomIndex(this.Parts.length - 2);
                let count = Math.ceil(RandomIndex(this.Parts.length - 2 - index) * this.mutationRate * 2) || 1;
                let parts = this.Parts.splice(index, count);
                this.Parts.push(...parts);
                this.Parts[index].Mutate();
            }
            else
                for (let i = 0; i < this.Parts.length; i++)
                {
                    let rand = Math.random();
                    if (rand < this.mutationRate)
                    {
                        this.Parts[i].Mutate();
                        i += 5;
                    }
                }
        }

        if (this.mutationRate > 0.2)
            this.mutationRate -= 0.004;
        return this;
    }

    Containers: Container[];
    HoleContainers: Container[];
    /**
     * 评估健康程度
     */
    Evaluate(bin: Path, bestCount: number, greedy = false, type?: PlaceType)
    {
        let parts = this.Parts;
        this.Containers = [];
        this.HoleContainers = [];
        for (let part of parts)
        {
            for (let i = 0; i < part.Holes.length; i++)
            {
                let hole = part.Holes[i];
                let container = new Container(hole.Contour);
                container.ParentId = part.Id;
                container.ChildrenIndex = i;
                container.ParentM = hole.OrigionMinPoint;
                this.HoleContainers.push(container);
            }
        }

        //网洞优先
        parts = parts.filter(p =>
        {
            for (let hole of this.HoleContainers)
            {
                if (hole.PutPart(p))
                    return false;
            }
            return true;
        });

        while (parts.length > 0)
        {
            if (this.Containers.length > bestCount)
            {
                this.Fitness = Math.ceil(bestCount) + 1;
                return;
            }
            let container = new Container(bin);
            container.ParentId = -1;
            if (greedy)
                container.PlaceType = type;
            else
                container.PlaceType = RandomIndex(2);
            if (this.mutationRate > 0.5)
            {
                let area = 0;
                let maxP: Part;
                for (let i = 0; i < parts.length; i++)
                {
                    let p = parts[i];
                    if (p.State.Contour.Area > area)
                    {
                        maxP = parts[i];
                        area = p.State.Contour.Area;
                    }
                }
                if (container.PutPart(maxP, greedy))
                    parts = parts.filter(p => p !== maxP);
            }
            parts = parts.filter(p =>
            {
                return container.PutPart(p, greedy) !== true;
            });
            if (!greedy)
            {
                parts = parts.filter(p =>
                {
                    return container.PutPart(p, true) !== true;
                });
            }
            this.Containers.push(container);
        }
        this.Fitness = this.Containers.length - 1 + arrayLast(this.Containers).UseRatio;
    }

    //#region -------------------------File-------------------------

    //对象从文件中读取数据,初始化自身
    ReadFile(file: NestFiler)
    {
        this.Fitness = file.Read();
        let count = file.Read() as number;
        this.Containers = [];
        for (let i = 0; i < count; i++)
            this.Containers.push(new Container().ReadFile(file, this.Parts));

        count = file.Read();
        this.HoleContainers = [];
        for (let i = 0; i < count; i++)
            this.HoleContainers.push(new Container().ReadFile(file, this.Parts));
    }

    //对象将自身数据写入到文件.
    WriteFile(f: NestFiler)
    {
        f.Write(this.Fitness);
        f.Write(this.Containers.length);
        for (let c of this.Containers)
            c.WriteFile(f);

        f.Write(this.HoleContainers.length);
        for (let c of this.HoleContainers)
            c.WriteFile(f);
    }
    //#endregion
}
