import { NestFiler } from "../Common/Filer";
import { PartState } from "./PartState";
import { Path } from "./Path";
import { PathGeneratorSingle } from "./PathGenerator";
import { Point } from "../Common/Point";
import { RandomIndex } from "../Common/Random";
import { Vector2 } from "../Common/Vector2";

/**
 * 零件类
 *  零件类可以绑定数据，也存在位置和旋转状态的信息
 *
 * 初始化零件：
 *  传入零件的轮廓，刀半径，包围容器（或者为空？）
 *  初始化用于放置的轮廓。将轮廓首点移动到0点，记录移动的点P。
 *
 * 零件放置位置：
 *  表示零件轮廓首点的位置。
 *
 * 零件的旋转：
 *  表示零件轮廓按照首点（0）旋转。
 *
 * 还原零件的放置状态：
 *  同样将零件移动到0点
 *  同样将零件旋转
 *  同样将零件移动到指定的位置
 *  零件可能处于容器中，变换到容器坐标系
 *
 */
export class Part<T = any, Matrix = any>
{
    Id: number;//用于确定Part的唯一性,并且有助于在其他Work中还原

    //下面数据只用于还原零件的显示状态,在优化中无作用
    UserData: T; //应该也是可序列化的实体
    PlaceCS: Matrix;//放置矩阵 Matrix4
    Parent: Part;

    Holes: PartState[] = [];
    //零件的不同旋转状态,这个数组会在所有的状态间共享,以便快速切换状态
    RotatedStates: PartState[] = [];
    PlacePosition: Point; //放置位置(相对于容器的位置)
    StateIndex = 0;       //放置状态

    get State(): PartState  //零件当前的状态
    {
        return this.RotatedStates[this.StateIndex];
    }

    //初始化零件的各个状态,按360度旋转个数
    Init(path: Path, bin: Path, rotateCount = 4): this
    {
        let rotations: number[] = [];
        let a = 2 * Math.PI / rotateCount;
        for (let i = 0; i < rotateCount; i++)
        {
            rotations.push(a * i);
        }
        this.Init2(path, bin, rotations);
        return this;
    }

    //初始化零件的各个状态,按旋转角度表
    Init2(path: Path, bin: Path, rotations: number[] = []): this
    {
        let pathP = path.OrigionMinPoint;
        let path_0 = PathGeneratorSingle.Allocate(path);
        let pathSet = new Set<Path>();

        //初始化零件的状态集合
        for (let pa of rotations)
        {
            let partState = new PartState();
            partState.Rotation = pa;
            if (pa === 0)
            {
                partState.Contour = path_0;
                partState.OrigionMinPoint = pathP;
            }
            else
            {
                let path_r = new Path(path.Points, pa);
                partState.Contour = PathGeneratorSingle.Allocate(path_r);
                partState.Contour.Area = path_0.Area;
                //避免重复的Path进入State
                if (pathSet.has(partState.Contour)) continue;
                let p0 = path_r.OrigionMinPoint;
                let c = Math.cos(-pa);
                let s = Math.sin(-pa);
                let x1 = p0.x * c - p0.y * s;
                let y1 = p0.x * s + p0.y * c;
                partState.OrigionMinPoint = new Vector2(pathP.x + x1, pathP.y + y1);
                if (partState.Contour.Origion === undefined)
                {
                    partState.Contour.Origion = path_0;
                    partState.Contour.OrigionMinPoint = p0;
                }
            }
            //记录已有Path
            pathSet.add(partState.Contour);
            //必须能放置
            if (bin.GetInsideNFP(partState.Contour))
            {
                this.RotatedStates.push(partState);
                PathGeneratorSingle.RegisterId(partState.Contour);
            }
        }

        //为了复用NFP,不管第0个Path是否可用,都注册它.
        if (this.RotatedStates.length > 4)
            PathGeneratorSingle.RegisterId(path_0);
        return this;
    }

    AppendHole(path: Path)
    {
        let hole = new PartState();
        hole.Contour = PathGeneratorSingle.Allocate(path);
        PathGeneratorSingle.RegisterId(hole.Contour);
        hole.OrigionMinPoint = path.OrigionMinPoint;
        hole.Rotation = 0;
        this.Holes.push(hole);
    }

    //浅克隆
    Clone()
    {
        let part = new Part();
        part.Id = this.Id;
        part.UserData = this.UserData;
        part.RotatedStates = this.RotatedStates;
        part.StateIndex = this.StateIndex;
        part.Holes = this.Holes;
        return part;
    }

    //旋转起来,改变自身旋转状态(变异)
    Mutate(): this
    {
        this.StateIndex = RandomIndex(this.RotatedStates.length, this.StateIndex);
        return this;
    }

    //#region -------------------------File-------------------------
    ReadFile(file: NestFiler)
    {
        this.Id = file.Read();
        let count = file.Read() as number;
        this.RotatedStates = [];
        for (let i = 0; i < count; i++)
        {
            let state = new PartState();
            state.ReadFile(file);
            this.RotatedStates.push(state);
        }

        count = file.Read();
        this.Holes = [];
        for (let i = 0; i < count; i++)
        {
            let state = new PartState();
            state.ReadFile(file);
            this.Holes.push(state);
        }
    }
    WriteFile(file: NestFiler)
    {
        file.Write(this.Id);
        file.Write(this.RotatedStates.length);
        for (let state of this.RotatedStates)
            state.WriteFile(file);

        file.Write(this.Holes.length);
        for (let hole of this.Holes)
            hole.WriteFile(file);
    }
    //#endregion
}
